@@ -54,10 +54,10 @@
  * SUCH DAMAGE.
  */
 
-#ifndef lint
+#if 0
 static char copyright[] =
 "@(#) Copyright (c) Apple Computer, Inc. 1997\n\n";
-#endif /* not lint */
+#endif
 
 /*
  * login [ name ]
@@ -116,13 +116,17 @@ void	 timedout __P((int));
 #ifdef KERBEROS
 int	 klogin __P((struct passwd *, char *, char *, char *));
 #endif
-void 	au_success();
-void 	au_fail(char *, int);
+void  au_success();
+void  au_fail(char *, int);
 
 
 extern void login __P((struct utmp *));
+static void bail(int, int);
+static void refused(const char *, const char *, int);
 
 #define	TTYGRPNAME	"tty"		/* name of group to own ttys */
+#define NO_SLEEP_EXIT	0
+#define SLEEP_EXIT	5
 
 /*
  * This bounds the time given to login.  Not a define so it can
@@ -140,6 +144,15 @@ int	authok;
 struct	passwd *pwd;
 int	failures;
 char	term[64], *hostname, *username = NULL, *tty;
+#ifdef USE_PAM
+static pam_handle_t *pamh = NULL;
+static struct pam_conv conv = { misc_conv, NULL };
+static int pam_err;
+static int pam_silent = PAM_SILENT;
+static int pam_cred_established;
+static int pam_session_established;
+#endif /* USE_PAM */
+int hflag;
 
 #define NA_EVENT_STR_SIZE 25
 au_tid_t tid;
@@ -152,20 +165,20 @@ main(argc, argv)
 	extern char **environ;
 	struct group *gr;
 	struct stat st;
-	struct timeval tp;
 	struct utmp utmp;
-	int ask, ch, cnt, oflag = 0, fflag, hflag, pflag, quietlog, rootlogin = 0, rval;
+	int ask, ch, cnt, oflag = 0, fflag, pflag, quietlog, rootlogin = 0;
 	uid_t uid;
 	uid_t euid;
 	gid_t egid;
-	char *domain, *p, *salt, *ttyn;
+	char *domain, *p, *ttyn;
 	char tbuf[MAXPATHLEN + 2], tname[sizeof(_PATH_TTY) + 10];
 	char localhost[MAXHOSTNAMELEN];
 #ifdef USE_PAM
-	pam_handle_t *pamh = NULL;
-	struct pam_conv conv = { misc_conv, NULL };
 	char **pmenv;
 	pid_t pid;
+#else
+	int rval;
+	char *salt;
 #endif
 
 	char auditsuccess = 1;
@@ -178,7 +191,6 @@ main(argc, argv)
 
 	openlog("login", LOG_ODELAY, LOG_AUTH);
 
-
 	/*
 	 * -p is used by getty to tell login not to destroy the environment
 	 * -f is used to skip a second login authentication
@@ -260,29 +272,29 @@ main(argc, argv)
 	}
 
 #ifdef USE_PAM
-	rval = pam_start("login", username, &conv, &pamh);
-	if( rval != PAM_SUCCESS ) {
-		fprintf(stderr, "login: PAM Error:  %s\n", pam_strerror(pamh, rval));
+	pam_err = pam_start("login", username, &conv, &pamh);
+	if( pam_err != PAM_SUCCESS ) {
+		fprintf(stderr, "login: PAM Error:  %s\n", pam_strerror(pamh, pam_err));
 		au_fail("PAM Error", 1);
 		exit(1);
 	}
-	rval = pam_set_item(pamh, PAM_TTY, tty);
-	if( rval != PAM_SUCCESS ) {
-		fprintf(stderr, "login: PAM Error: %s\n", pam_strerror(pamh, rval));
+	pam_err = pam_set_item(pamh, PAM_TTY, tty);
+	if( pam_err != PAM_SUCCESS ) {
+		fprintf(stderr, "login: PAM Error: %s\n", pam_strerror(pamh, pam_err));
 		au_fail("PAM Error", 1);
 		exit(1);
 	}
 
-	rval = pam_set_item(pamh, PAM_RHOST, hostname);
-	if( rval != PAM_SUCCESS ) {
-		fprintf(stderr, "login: PAM Error: %s\n", pam_strerror(pamh, rval));
+	pam_err = pam_set_item(pamh, PAM_RHOST, hostname);
+	if( pam_err != PAM_SUCCESS ) {
+		fprintf(stderr, "login: PAM Error: %s\n", pam_strerror(pamh, pam_err));
 		au_fail("PAM Error", 1);
 		exit(1);
 	}
 
-	rval = pam_set_item(pamh, PAM_USER_PROMPT, "login: ");
-	if( rval != PAM_SUCCESS ) {
-		fprintf(stderr, "login: PAM Error: %s\n", pam_strerror(pamh, rval));
+	pam_err = pam_set_item(pamh, PAM_USER_PROMPT, "login: ");
+	if( pam_err != PAM_SUCCESS ) {
+		fprintf(stderr, "login: PAM Error: %s\n", pam_strerror(pamh, pam_err));
 		au_fail("PAM Error", 1);
 		exit(1);
 	}
@@ -295,15 +307,15 @@ main(argc, argv)
 		rootlogin = 1;
 
 	if( (pwd != NULL) && fflag && ((uid == 0) || (uid == pwd->pw_uid)) ){
-		rval = 0;
+		pam_err = 0;
 		auditsuccess = 0; /* we've simply opened a terminal window */
 	} else {
 
-		rval = pam_authenticate(pamh, 0);
-		while( (!oflag) && (cnt++ < 10) && ((rval == PAM_AUTH_ERR) ||
-				(rval == PAM_USER_UNKNOWN) ||
-				(rval == PAM_CRED_INSUFFICIENT) ||
-				(rval == PAM_AUTHINFO_UNAVAIL))) {
+		pam_err = pam_authenticate(pamh, 0);
+		while( (!oflag) && (cnt++ < 10) && ((pam_err == PAM_AUTH_ERR) ||
+				(pam_err == PAM_USER_UNKNOWN) ||
+				(pam_err == PAM_CRED_INSUFFICIENT) ||
+				(pam_err == PAM_AUTHINFO_UNAVAIL))) {
 			/* 
 			 * we are not exiting here, but this corresponds to 
 		 	 * a failed login event, so set exitstatus to 1 
@@ -317,10 +329,10 @@ main(argc, argv)
 			if( (pwd != NULL) && (pwd->pw_uid == 0) )
 				rootlogin = 1;
 			pam_set_item(pamh, PAM_USER, username);
-			rval = pam_authenticate(pamh, 0);
+			pam_err = pam_authenticate(pamh, 0);
 		}
 
-		if( rval != PAM_SUCCESS ) {
+		if( pam_err != PAM_SUCCESS ) {
 			pam_get_item(pamh, PAM_USER, (void *)&username);
 			badlogin(username);
 			printf("Login incorrect\n");
@@ -328,32 +340,31 @@ main(argc, argv)
 			exit(1);
 		}
 
-		rval = pam_acct_mgmt(pamh, 0);
-		if( rval == PAM_NEW_AUTHTOK_REQD ) {
-			rval = pam_chauthtok(pamh, PAM_CHANGE_EXPIRED_AUTHTOK);
+		pam_err = pam_acct_mgmt(pamh, 0);
+		if( pam_err == PAM_NEW_AUTHTOK_REQD ) {
+			pam_err = pam_chauthtok(pamh, PAM_CHANGE_EXPIRED_AUTHTOK);
 		}
-		if( rval != PAM_SUCCESS ) {
-			fprintf(stderr, "login: PAM Error: %s\n", pam_strerror(pamh, rval));
+		if( pam_err != PAM_SUCCESS ) {
+			fprintf(stderr, "login: PAM Error: %s\n", pam_strerror(pamh, pam_err));
 			au_fail("PAM error", 1);
 			exit(1);
 		}
-
 	}
 
-	rval = pam_get_item(pamh, PAM_USER, (void *)&username);
-	if( (rval == PAM_SUCCESS) && username && *username) 
+	pam_err = pam_get_item(pamh, PAM_USER, (void *)&username);
+	if( (pam_err == PAM_SUCCESS) && username && *username) 
 		pwd = getpwnam(username);
 
-	rval = pam_open_session(pamh, 0);
-	if( rval != PAM_SUCCESS ) {
-		fprintf(stderr, "login: PAM Error: %s\n", pam_strerror(pamh, rval));
+	pam_err = pam_open_session(pamh, 0);
+	if( pam_err != PAM_SUCCESS ) {
+		fprintf(stderr, "login: PAM Error: %s\n", pam_strerror(pamh, pam_err));
 		au_fail("PAM error", 1);
 		exit(1);
 	}
 
-	rval = pam_setcred(pamh, PAM_ESTABLISH_CRED);
-	if( rval != PAM_SUCCESS ) {
-		fprintf(stderr, "login: PAM Error: %s\n", pam_strerror(pamh, rval));
+	pam_err = pam_setcred(pamh, PAM_ESTABLISH_CRED);
+	if( pam_err != PAM_SUCCESS ) {
+		fprintf(stderr, "login: PAM Error: %s\n", pam_strerror(pamh, pam_err));
 		au_fail("PAM error", 1);
 		exit(1);
 	}
@@ -485,25 +496,27 @@ main(argc, argv)
 	setegid(pwd->pw_gid);
 	seteuid(rootlogin ? 0 : pwd->pw_uid);
 
-	/* First do a stat in case the homedir is automounted */
-	stat(pwd->pw_dir,&st);
+      	/* First do a stat in case the homedir is automounted */
+      	stat(pwd->pw_dir,&st);
 
-	if (chdir(pwd->pw_dir) < 0) {
-		(void)printf("No home directory %s!\n", pwd->pw_dir);
+	if (!*pwd->pw_dir || chdir(pwd->pw_dir) < 0) {
+		printf("No home directory %s!\n", pwd->pw_dir);
 		if (chdir("/")) {
-			exit(0);
+			refused("Cannot find root directory", "HOMEDIR", 1);
 		}
-		pwd->pw_dir = "/";
-		(void)printf("Logging in with home = \"/\".\n");
+		pwd->pw_dir = strdup("/");
+		if (pwd->pw_dir == NULL) {
+			syslog(LOG_NOTICE, "strdup(): %m");
+			bail(SLEEP_EXIT, 1);
+		}
+		printf("Logging in with home = \"/\".\n");
 	}
-
 	seteuid(euid);
 	setegid(egid);
 
 	quietlog = access(_PATH_HUSHLOGIN, F_OK) == 0;
 
 	/* Nothing else left to fail -- really log in. */
-
 	memset((void *)&utmp, 0, sizeof(utmp));
 	(void)time(&utmp.ut_time);
 	(void)strncpy(utmp.ut_name, username, sizeof(utmp.ut_name));
@@ -519,10 +532,15 @@ main(argc, argv)
 	(void)chmod(ttyn, 0620);
 	(void)setgid(pwd->pw_gid);
 
-	initgroups(username, pwd->pw_gid);
+	if (initgroups(username, pwd->pw_gid) == -1)
+		syslog(LOG_ERR, "login: initgroups() failed");
 
 	if (*pwd->pw_shell == '\0')
-		pwd->pw_shell = _PATH_BSHELL;
+		pwd->pw_shell = strdup(_PATH_BSHELL);
+        if (pwd->pw_shell == NULL) {
+                syslog(LOG_NOTICE, "strdup(): %m");
+                bail(SLEEP_EXIT, 1);
+        }
 
 	/* Destroy environment unless user has requested its preservation. */
 	if (!pflag) {
@@ -553,8 +571,10 @@ main(argc, argv)
 	} else if( pid != 0 ) {
 		waitpid(pid, NULL, 0);
 		pam_setcred(pamh, PAM_DELETE_CRED);
-		rval = pam_close_session(pamh, 0);
-		pam_end(pamh,rval);
+		pam_err = pam_close_session(pamh, 0);
+		pam_end(pamh,pam_err);
+		chown(ttyn, 0, 0);
+		chmod(ttyn, 0666);
 		exit(0);
 	}
 
@@ -603,7 +623,6 @@ main(argc, argv)
 	else
 		(void) setuid(pwd->pw_uid);
 
-	
 	execlp(pwd->pw_shell, tbuf, 0);
 	err(1, "%s", pwd->pw_shell);
 }
@@ -836,7 +855,6 @@ checknologin()
 	if ((fd = open(_PATH_NOLOGIN, O_RDONLY, 0)) >= 0) {
 		while ((nchars = read(fd, tbuf, sizeof(tbuf))) > 0)
 			(void)write(fileno(stdout), tbuf, nchars);
-
 		au_fail("No login", 0);
 		sleepexit(0);
 	}
@@ -929,3 +947,66 @@ sleepexit(eval)
 	(void)sleep(5);
 	exit(eval);
 }
+
+static void
+refused(const char *msg, const char *rtype, int lout)
+{
+
+        if (msg != NULL)
+		printf("%s.\n", msg);
+        if (hflag)
+                syslog(LOG_NOTICE, "LOGIN %s REFUSED (%s) FROM %s ON TTY %s",
+		       pwd->pw_name, rtype, hostname, tty);
+        else
+                syslog(LOG_NOTICE, "LOGIN %s REFUSED (%s) ON TTY %s",
+		       pwd->pw_name, rtype, tty);
+        if (lout)
+                bail(SLEEP_EXIT, 1);
+}
+
+#ifdef USE_PAM
+/*
+ * Log a PAM error
+ */
+static void
+pam_syslog(const char *msg)
+{
+        syslog(LOG_ERR, "%s: %s", msg, pam_strerror(pamh, pam_err));
+}
+
+/*
+ * Shut down PAM
+ */
+static void
+pam_cleanup()
+{
+
+        if (pamh != NULL) {
+                if (pam_session_established) {
+                        pam_err = pam_close_session(pamh, 0);
+                        if (pam_err != PAM_SUCCESS)
+                                pam_syslog("pam_close_session()");
+                }
+                pam_session_established = 0;
+                if (pam_cred_established) {
+                        pam_err = pam_setcred(pamh, pam_silent|PAM_DELETE_CRED);
+                        if (pam_err != PAM_SUCCESS)
+                                pam_syslog("pam_setcred()");
+                }
+                pam_cred_established = 0;
+                pam_end(pamh, pam_err);
+                pamh = NULL;
+        }
+}
+#endif /* USE_PAM */
+/*
+ * Exit, optionally after sleeping a few seconds
+ */
+void
+bail(int sec, int eval)
+{
+
+        pam_cleanup();
+        (void)sleep(sec);
+        exit(eval);
+}
